import abc

from magma.debug import debug_wire


class MagmaProtocolMeta(type):
    @abc.abstractmethod
    def _to_magma_(cls):
        # To retrieve underlying magma type.
        raise NotImplementedError()

    @abc.abstractmethod
    def _qualify_magma_(cls, direction: 'Direction'):
        # To qualify underlying type (e.g. give me a Foo with the underlying
        # type qualified to be an input).
        raise NotImplementedError()

    @abc.abstractmethod
    def _flip_magma_(cls):
        # To flip underlying type (e.g. give me a Foo with the underlying type
        # flipped).
        raise NotImplementedError()

    def qualify(cls, direction: 'Direction'):
        return cls._qualify_magma_(direction)

    @abc.abstractmethod
    def _from_magma_value_(cls, val: 'Type'):
        # To create an instance from a value.
        raise NotImplementedError()

    def _is_oriented_magma_(cls, direction):
        return cls._to_magma_().is_oriented(direction)

    def is_oriented(cls, direction):
        return cls._is_oriented_magma_(direction)

    def flip(cls):
        return cls._flip_magma_()

    def __len__(cls):
        return len(cls._to_magma_())

    def is_wireable(cls, rhs):
        return cls._to_magma_().is_wireable(rhs)

    def is_bindable(cls, rhs):
        return cls._to_magma_().is_bindable(rhs)

    @property
    def direction(cls) -> int:
        return cls._to_magma_().direction


class MagmaProtocol(metaclass=MagmaProtocolMeta):
    @abc.abstractmethod
    def _get_magma_value_(self):
        # To access underlying magma value.
        raise NotImplementedError()

    @classmethod
    def is_clock(cls):
        return cls._to_magma_().is_clock()

    @classmethod
    def is_inout(cls):
        return cls._to_magma_().is_inout()

    @classmethod
    def is_input(cls):
        return cls._to_magma_().is_input()

    @classmethod
    def is_mixed(cls):
        return cls._to_magma_().is_mixed()

    @classmethod
    def is_output(cls):
        return cls._to_magma_().is_output()

    def iswhole(self):
        return self._get_magma_value_().iswhole()

    def value(self):
        return self._get_magma_value_().value()

    def trace(self):
        return self._get_magma_value_().trace()

    def driven(self):
        return self._get_magma_value_().driven()

    def flatten(self):
        return self._get_magma_value_().flatten()

    def __iter__(self):
        return iter(self._get_magma_value_())

    @property
    def name(self):
        return self._get_magma_value_().name

    @debug_wire
    def wire(self, other, debug_info):
        if isinstance(other, MagmaProtocol):
            other = other._get_magma_value_()
        self._get_magma_value_().wire(other, debug_info)

    def unwire(self, other):
        if isinstance(other, MagmaProtocol):
            other = other._get_magma_value_()
        self._get_magma_value_().unwire(other)

    def __imatmul__(self, other):
        self.wire(other)
        return self

    def connection_iter(self):
        return self._get_magma_value_().connection_iter()


def magma_type(T):
    if issubclass(T, MagmaProtocol):
        return T._to_magma_()
    return T


def magma_value(value):
    if isinstance(value, MagmaProtocol):
        return value._get_magma_value_()
    return value



from magma.array import Array
from magma.circuit import Circuit
from magma.clock import AsyncReset, AsyncResetN, Clock, is_clock_or_nested_clock
from magma.conversions import convertbit
from magma.interface import IO
from magma.passes import DefinitionPass, pass_lambda
from magma.t import In, Out, Direction
from magma.tuple import Tuple
from magma.wire import wire


_NAMED_TYPES = (AsyncReset, AsyncResetN, Clock)


class InsertWrapCasts(DefinitionPass):
    def define_wrap(self, wrap_type, in_type, out_type):
        class Wrap(Circuit):
            name = f"coreir_wrap{wrap_type}".replace("(", "").replace(")", "")
            io = IO(**{"in": In(in_type), "out": Out(out_type)})
            coreir_genargs = {"type": wrap_type}
            coreir_name = "wrap"
            coreir_lib = "coreir"
            primitive = True

            def simulate(self, value_store, state_store):
                input_val = value_store.get_value(getattr(self, "in"))
                value_store.set_value(self.out, input_val)
        return Wrap

    def _recurse(self, port, definition):
        wrapped = False
        for t in port:
            wrapped |= self.wrap_if_named_type(t, definition)
        return wrapped

    def wrap_if_named_type(self, port, definition):
        if port.is_mixed():
            return self._recurse(port, definition)
        if not port.driven():
            return False
        value = port.value()
        if isinstance(port, Tuple):
            return self._recurse(port, definition)
        if isinstance(port, Array):
            # Avoid recursion when possible (for Array2) by checking the nested
            # array type and only descending if necessary
            # Note(leonardt): If value is anon, we need to check the children
            # via recursion in case the children are named types (since
            # .value() will return Array[N, T.flip()], the anon value may not
            # have the namedtypes in its type)
            if (is_clock_or_nested_clock(type(port), _NAMED_TYPES) or
                    is_clock_or_nested_clock(type(value), _NAMED_TYPES) or
                    value.anon()):
                for child, _ in port.connection_iter():
                    if not self.wrap_if_named_type(child, definition):
                        return False
                return True
            else:
                return self.wrap_if_named_type(value, definition)
        if not (isinstance(port, _NAMED_TYPES) or
                isinstance(value, _NAMED_TYPES)):
            return self.wrap_if_named_type(value, definition)
        undirected_t = type(port).qualify(Direction.Undirected)
        if issubclass(type(value), undirected_t):
            return self.wrap_if_named_type(value, definition)
        port_is_clock_type = isinstance(port, (AsyncReset, AsyncResetN, Clock))
        if port_is_clock_type:
            T = Out(type(port))
        else:
            T = In(type(value))

        with definition.open():
            port.unwire(value)
            in_type, out_type = type(port), type(value)
            if not port_is_clock_type:
                # value is the clock type, flip wrap types
                out_type, in_type = in_type, out_type
            inst = self.define_wrap(T, in_type, out_type)()
            wire(convertbit(value, in_type), getattr(inst, "in"))
            wire(inst.out, convertbit(port, out_type))
        return True

    def __call__(self, definition):
        # copy, because wrapping might add instances
        instances = definition.instances[:]
        for instance in instances:
            if type(instance).coreir_name == "wrap" or \
                    type(instance).coreir_name == "unwrap":
                continue
            for port in instance.interface.ports.values():
                self.wrap_if_named_type(port, definition)
        for port in definition.interface.ports.values():
            self.wrap_if_named_type(port, definition)


insert_wrap_casts = pass_lambda(InsertWrapCasts)



import abc
import json
from typing import Any, Iterable, Mapping, Tuple

from magma.common import make_delegator_cls


class _Sentinel:
    _instances = []

    def __init__(self, string):
        _Sentinel._instances.append(self)
        self._string = string

    @classmethod
    def match(cls, string):
        for sentinel in cls._instances:
            if sentinel.string == string:
                return sentinel
        return None

    @property
    def string(self):
        return self._string

    def __repr__(self):
        return f"_Sentinel({self._string})"

    def __str__(self):
        return self._string


SYMBOL_TABLE_INLINED_INSTANCE = _Sentinel("__SYMBOL_TABLE_INLINED_INSTANCE__")
SYMBOL_TABLE_EMPTY = _Sentinel("__SYMBOL_TABLE_EMPTY__")
InstanceNameType = Tuple[_Sentinel, str]


def _is_tuple_annotation(T):
    try:
        origin = T.__origin__
    except AttributeError:
        return False
    return origin is tuple


def _unwrap_value(value):
    if isinstance(value, tuple):
        return tuple(_unwrap_value(v) for v in value)
    if not isinstance(value, _Sentinel):
        return value
    return value.string


def _try_isinstance(value, T):
    try:
        ret = isinstance(value, T)
    except Exception:
        return None
    return ret


def _type_check_value(value, T):
    sentinel_match = _Sentinel.match(value)
    if sentinel_match is not None:
        value = sentinel_match
    if _try_isinstance(value, T) is True:
        return value
    return None


class SymbolTableInterface(abc.ABC):
    @abc.abstractmethod
    def get_module_name(self, in_module_name: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_instance_name(self,
                          in_module_name: str,
                          in_instance_name: str) -> InstanceNameType:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_port_name(self,
                      in_module_name: str,
                      in_port_name: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_inlined_instance_name(self,
                                  in_module_name: str,
                                  in_parent_instance_name: str,
                                  in_child_instance_name: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_instance_type(self,
                          in_module_name: str,
                          in_instance_name: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_module_name(self,
                        in_module_name: str,
                        out_module_name: str) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_instance_name(self,
                          in_module_name: str,
                          in_instance_name: str,
                          out_instance_name: InstanceNameType) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_port_name(self,
                      in_module_name: str,
                      in_port_name: str,
                      out_port_name: str) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_inlined_instance_name(self,
                                  in_module_name: str,
                                  in_parent_instance_name: str,
                                  in_child_instance_name: str,
                                  out_instance_name: InstanceNameType) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_instance_type(self,
                          in_module_name: str,
                          in_instance_name: str,
                          out_type: str) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def module_names(self) -> Mapping[str, str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def instance_names(self) -> Mapping[Tuple[str, str], InstanceNameType]:
        raise NotImplementedError()

    @abc.abstractmethod
    def port_names(self) -> Mapping[Tuple[str, str], str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def inlined_instance_names(self) -> Mapping[Tuple[str, str, str],
                                                InstanceNameType]:
        raise NotImplementedError()

    @abc.abstractmethod
    def instance_types(self) -> Mapping[Tuple[str, str], str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def as_dict(self, for_json):
        raise NotImplementedError()

    @abc.abstractmethod
    def as_json(self, **kwargs):
        raise NotImplementedError()


class ImmutableSymbolTable(SymbolTableInterface):
    def set_module_name(self,
                        in_module_name: str,
                        out_module_name: str) -> None:
        raise Exception("Can not set an immutable symbol table")

    def set_instance_name(self,
                          in_module_name: str,
                          in_instance_name: str,
                          out_instance_name: InstanceNameType) -> None:
        raise Exception("Can not set an immutable symbol table")

    def set_port_name(self,
                      in_module_name: str,
                      in_port_name: str,
                      out_port_name: str) -> None:
        raise Exception("Can not set an immutable symbol table")

    def set_inlined_instance_name(self,
                                  in_module_name: str,
                                  in_parent_instance_name: str,
                                  in_child_instance_name: str,
                                  out_instance_name: InstanceNameType) -> None:
        raise Exception("Can not set an immutable symbol table")

    def set_instance_type(self,
                          in_module_name: str,
                          in_instance_name: str,
                          out_type: str) -> None:
        raise Exception("Can not set an immutable symbol table")


DelegatorSymbolTable = make_delegator_cls(SymbolTableInterface)


class _FieldMeta(type):
    def __new__(metacls, name, bases, dct, info=(None, None), **kwargs):
        if "_info_" in dct:
            raise TypeError(
                "class attribute _info_ is reversed by the type machinery")
        dct['_info_'] = info
        return super().__new__(metacls, name, bases, dct, **kwargs)

    def __getitem__(cls, idx: Tuple[Any, Any]):
        metacls = type(cls)
        bases = tuple([cls])
        name = f"{cls.__name__}[{idx}]"
        t = metacls(name, bases, {}, info=(cls, idx))
        t.__module__ = cls.__module__
        return t

    @property
    def key_type(cls):
        return cls._info_[1][0]

    @property
    def key_length(cls):
        key_type = cls.key_type
        if isinstance(key_type, type):
            return 1
        assert isinstance(key_type, tuple)
        return len(key_type)

    @property
    def value_type(cls):
        return cls._info_[1][1]

    @property
    def value_length(cls):
        value_type = cls.value_type
        if isinstance(value_type, type):
            return 1
        assert _is_tuple_annotation(value_type)
        return len(value_type.__args__)

    def _parse_key(cls, key, from_json):
        if from_json:
            key = tuple(key.split(","))
            if len(key) == 1:
                key = key[0]
        if cls.key_length == 1:
            if not isinstance(key, cls.key_type):
                raise ValueError()
        else:
            if not isinstance(key, tuple):
                raise ValueError()
            if len(key) != cls.key_length:
                raise ValueError()
        return key

    def _parse_value(cls, value, from_json):
        if from_json:
            if isinstance(value, list):
                value = tuple(value)
        if cls.value_length == 1:
            value = _type_check_value(value, cls.value_type)
            if value is None:
                raise ValueError()
            return value
        if not isinstance(value, tuple):
            raise ValueError()
        if len(value) != cls.value_length:
            raise ValueError()
        value = tuple(_type_check_value(v, T)
                      for v, T in zip(value, cls.value_type.__args__))
        if any(v is None for v in value):
            raise ValueError()
        return value

    def from_dict(cls, dct, from_json=False):
        ret = cls()
        for k, v in dct.items():
            k, v = cls._parse_key(k, from_json), cls._parse_value(v, from_json)
            ret.set(k, v)
        return ret


class _Field(object, metaclass=_FieldMeta):
    def __init__(self):
        self._map = {}

    def set(self, key, value):
        try:
            self._map[key]
            raise ValueError(f"key {key} already mapped")
        except KeyError:
            pass
        self._map[key] = value

    def get(self, key):
        return self._map[key]

    def dct(self, for_json=False):
        cls = type(self)
        if not for_json or cls.key_length == 1:
            return self._map.copy()
        return {",".join([str(ki) for ki in k]): _unwrap_value(v)
                for k, v in self._map.items()}


class SymbolTable(SymbolTableInterface):
    _fields_ = {
        "module_names": _Field[str, str],
        "instance_names": _Field[(str, str), InstanceNameType],
        "port_names": _Field[(str, str), str],
        "inlined_instance_names": _Field[(str, str, str), InstanceNameType],
        "instance_types": _Field[(str, str), str],
    }

    def __init__(self):
        self._mappings = {k: T() for k, T, in SymbolTable._fields_.items()}

    def get_module_name(self, in_module_name: str) -> str:
        return self._mappings["module_names"].get(in_module_name)

    def get_instance_name(self,
                          in_module_name: str,
                          in_instance_name: str) -> InstanceNameType:
        key = in_module_name, in_instance_name
        return self._mappings["instance_names"].get(key)

    def get_port_name(self,
                      in_module_name: str,
                      in_port_name: str) -> str:
        key = in_module_name, in_port_name
        return self._mappings["port_names"].get(key)

    def get_inlined_instance_name(
            self,
            in_module_name: str,
            in_parent_instance_name: str,
            in_child_instance_name: str) -> InstanceNameType:
        key = in_module_name, in_parent_instance_name, in_child_instance_name
        return self._mappings["inlined_instance_names"].get(key)

    def get_instance_type(self,
                          in_module_name: str,
                          in_instance_name: str) -> str:
        key = in_module_name, in_instance_name
        return self._mappings["instance_types"].get(key)

    def set_module_name(self,
                        in_module_name: str,
                        out_module_name: str) -> None:
        self._mappings["module_names"].set(in_module_name, out_module_name)

    def set_instance_name(self,
                          in_module_name: str,
                          in_instance_name: str,
                          out_instance_name: InstanceNameType) -> None:
        key = in_module_name, in_instance_name
        self._mappings["instance_names"].set(key, out_instance_name)

    def set_port_name(self,
                      in_module_name: str,
                      in_port_name: str,
                      out_port_name: str) -> None:
        key = in_module_name, in_port_name
        self._mappings["port_names"].set(key, out_port_name)

    def set_inlined_instance_name(self,
                                  in_module_name: str,
                                  in_parent_instance_name: str,
                                  in_child_instance_name: str,
                                  out_instance_name: InstanceNameType) -> None:
        key = in_module_name, in_parent_instance_name, in_child_instance_name
        self._mappings["inlined_instance_names"].set(key, out_instance_name)

    def set_instance_type(self,
                          in_module_name: str,
                          in_instance_name: str,
                          out_type: str) -> None:
        key = in_module_name, in_instance_name
        self._mappings["instance_types"].set(key, out_type)

    def module_names(self) -> Mapping[str, str]:
        return self._mappings["module_names"].dct()

    def instance_names(self) -> Mapping[Tuple[str, str], InstanceNameType]:
        return self._mappings["instance_names"].dct()

    def port_names(self) -> Mapping[Tuple[str, str], str]:
        return self._mappings["port_names"].dct()

    def inlined_instance_names(self) -> Mapping[Tuple[str, str],
                                                InstanceNameType]:
        return self._mappings["inlined_instance_names"].dct()

    def instance_types(self) -> Mapping[Tuple[str, str], str]:
        return self._mappings["instance_types"].dct()

    def as_dict(self, for_json=False):
        return {
            name: field.dct(for_json)
            for name, field in self._mappings.items()
        }

    @staticmethod
    def from_dict(dct, from_json=False, allow_missing=True):
        table = SymbolTable()
        for name, mapping in table._mappings.items():
            T = type(mapping)
            try:
                entry = dct[name]
            except KeyError as err:
                if not allow_missing:
                    raise  # re-raise original KeyError
            else:
                table._mappings[name] = T.from_dict(entry, from_json)
        return table

    def as_json(self, **kwargs):
        return json.dumps(self.as_dict(for_json=True), **kwargs)

    @staticmethod
    def from_json(value, allow_missing=True):
        dct = json.loads(value)
        return SymbolTable.from_dict(
            dct, from_json=True, allow_missing=allow_missing)



import itertools
from functools import lru_cache
from collections import OrderedDict
from hwtypes.adt import (
    TupleMeta,
    Tuple as Tuple_,
    AnonymousProduct,
    AnonymousProductMeta,
    Product,
    ProductMeta,
)
from hwtypes import BitVector, Bit
from hwtypes.adt_meta import BoundMeta, RESERVED_SUNDERS
from hwtypes.util import TypedProperty, OrderedFrozenDict
from .common import deprecated
from .ref import AnonRef, TupleRef
from .t import Type, Kind, Direction
from .compatibility import IntegerTypes
from .debug import debug_wire, get_callee_frame_info
from .logging import root_logger
from .protocol_type import magma_type, magma_value

from magma.wire_container import WiringLog
from magma.wire import wire
from magma.protocol_type import MagmaProtocol
from magma.operator_utils import output_only


_logger = root_logger()


#
# Create an Tuple
#
#  Tuple()
#  - creates a new tuple value
#  Tuple(v0, v1, ..., vn)
#  - creates a new tuple value where each field equals vi
#

class TupleKind(TupleMeta, Kind):
    def __new__(mcs, name, bases, namespace, fields=None, **kwargs):
        for rname in RESERVED_SUNDERS:
            if rname in namespace:
                raise ReservedNameError(f'class attribute {rname} is reserved by the type machinery')

        bound_types = fields
        has_bound_base = False
        unbound_bases = []
        for base in bases:
            if isinstance(base, BoundMeta):
                if base.is_bound:
                    has_bound_base = True
                    if bound_types is None:
                        bound_types = base.fields
                    else:
                        for x, y in zip(bound_types, base.fields):
                            if not (issubclass(x, y) or issubclass(y, x)):
                                try:
                                    if x.is_bindable(y):
                                        continue
                                except AttributeError:
                                    pass
                                try:
                                    if y.is_bindable(x):
                                        continue
                                except AttributeError:
                                    pass
                                raise TypeError(
                                    "Can't inherit from multiple different bound_types"
                                )
                else:
                    unbound_bases.append(base)

        t = type.__new__(mcs, name, bases, namespace, **kwargs)
        t._fields_ = bound_types
        t._unbound_base_ = None

        if bound_types is None:
            # t is a unbound type
            t._unbound_base_ = t
        elif len(unbound_bases) == 1:
            # t is constructed from an unbound type
            t._unbound_base_ = unbound_bases[0]
        elif not has_bound_base:
            # this shouldn't be reachable
            raise AssertionError("Unreachable code")

        return t

    def _from_idx(cls, idx):
        # Some of this should probably be in GetitemSyntax
        mcs = type(cls)

        try:
            return mcs._class_cache[cls, idx]
        except KeyError:
            pass

        if cls.is_bound:
            raise TypeError('Type is already bound')

        undirected_idx = tuple(v.qualify(Direction.Undirected) for v in idx)
        if any(x is not y for x, y in zip(undirected_idx, idx)):
            bases = [cls[undirected_idx]]
        else:
            bases = [cls]
        bases = tuple(bases)
        class_name = cls._name_cb(idx)

        t = mcs(class_name, bases, {'__module__' : cls.__module__}, fields=idx)
        if t._unbound_base_ is None:
            t._unbound_base_ = cls
        mcs._class_cache[cls, idx] = t
        t._cached_ = True
        return t

    def qualify(cls, direction):
        new_fields = []
        for T in cls.fields:
            new_fields.append(T.qualify(direction))
        return cls.unbound_t[new_fields]

    def flip(cls):
        new_fields = []
        for T in cls.fields:
            new_fields.append(T.flip())
        return cls.unbound_t[new_fields]

    @property
    @lru_cache()
    def direction(cls):
        directions = iter(t.direction for t in cls.fields)
        first = next(directions)
        if all(d == first for d in directions):
            return first
        return None

    def __call__(cls, *args, **kwargs):
        obj = cls.__new__(cls, *args)
        obj.__init__(*args, **kwargs)
        return obj

    @property
    def N(cls):
        return len(cls.fields)

    def __str__(cls):
        if not cls.is_bound:
            return cls.__name__
        s = "Tuple("
        s += ",".join(str(T) for T in cls.fields)
        s += ")"
        return s

    def is_wireable(cls, rhs):
        rhs = magma_type(rhs)
        if not isinstance(rhs, TupleKind) or len(cls.fields) != len(rhs.fields):
            return False
        for idx, T in enumerate(cls.fields):
            if not T.is_wireable(rhs[idx]):
                return False
        return True

    def is_bindable(cls, rhs):
        rhs = magma_type(rhs)
        if not isinstance(rhs, TupleKind) or len(cls.fields) != len(rhs.fields):
            return False
        for idx, T in enumerate(cls.fields):
            if not T.is_bindable(rhs[idx]):
                return False
        return True

    def __eq__(cls, rhs):
        if not isinstance(rhs, TupleKind):
            return False

        if not cls.is_bound:
            return not rhs.is_bound

        return cls.fields == rhs.fields

    __hash__ = TupleMeta.__hash__


class Tuple(Type, Tuple_, metaclass=TupleKind):
    def __init__(self, *largs, **kwargs):

        Type.__init__(self, **kwargs) # name=

        self.ts = []
        if len(largs) > 0:
            assert len(largs) == len(self)
            for k, t, T in zip(self.keys(), largs, self.types()):
                if isinstance(t, (IntegerTypes, BitVector, Bit)):
                    t = T(t)
                self.ts.append(t)
                if not isinstance(self, AnonProduct):
                    setattr(self, k, t)
        else:
            for k, T in zip(self.keys(), self.types()):
                ref = TupleRef(self, k)
                if issubclass(T, MagmaProtocol):
                    t = T._from_magma_value_(T._to_magma_()(name=ref))
                else:
                    t = T(name=ref)
                self.ts.append(t)
                if not isinstance(self, AnonProduct):
                    setattr(self, k, t)

    __hash__ = Type.__hash__

    @classmethod
    def is_oriented(cls, direction):
        return all(v.is_oriented(direction) for v in cls.fields)

    @classmethod
    def is_clock(cls):
        return False

    @classmethod
    @deprecated
    def isoriented(cls, direction):
        return cls.is_oriented(direction)

    @classmethod
    def keys(cls):
        return [str(i) for i in range(len(cls.fields))]

    @output_only("Cannot use == on an input")
    def __eq__(self, rhs):
        if not isinstance(rhs, type(self)):
            return NotImplemented
        else:
            return self.ts == rhs.ts

    @output_only("Cannot use != on an input")
    def __ne__(self, rhs):
        return ~(self == rhs)

    def __repr__(self):
        if not self.name.anon():
            return super().__repr__()
        ts = [repr(t) for t in self.ts]
        kts = ['{}={}'.format(k, v) for k, v in zip(self.keys(), ts)]
        return 'tuple(dict({})'.format(', '.join(kts))

    def __getitem__(self, key):
        if key in self.keys():
            key = list(self.keys()).index(key)
        return self.ts[key]

    def __setitem__(self, key, val):
        old = self[key]
        if old is not val:
            _logger.error(
                WiringLog(f"May not mutate tuple, trying to replace "
                          f"{{}}[{key}] ({{}}) with {{}}", self, old, val)
            )

    def __len__(self):
        return len(type(self).fields)

    @classmethod
    def flat_length(cls):
        return sum(magma_type(T).flat_length() for T in cls.types())

    def __call__(self, o):
        return self.wire(o, get_callee_frame_info())


    @debug_wire
    def wire(i, o, debug_info):
        o = magma_value(o)
        if not isinstance(o, Tuple):
            _logger.error(
                WiringLog(f"Cannot wire {{}} (type={type(o)}) to {{}} "
                          f"(type={type(i)}) because {{}} is not a Tuple",
                          o, i, o),
                debug_info=debug_info
            )
            return

        if i.keys() != o.keys():
            _logger.error(
                WiringLog(f"Cannot wire {{}} (type={type(o)}, "
                          f"keys={list(i.keys())}) to {{}} (type={type(i)}, "
                          f"keys={list(o.keys())}) because the tuples do not "
                          f"have the same keys", o, i),
                debug_info=debug_info
            )
            return

        for i_elem, o_elem in zip(i, o):
            i_elem = magma_value(i_elem)
            o_elem = magma_value(o_elem)
            wire(o_elem, i_elem, debug_info)

    def unwire(i, o):
        for i_elem, o_elem in zip(i, o):
            if o_elem.is_input():
                o_elem.unwire(i_elem)
            else:
                i_elem.unwire(o_elem)

    def driven(self):
        for t in self.ts:
            if not t.driven():
                return False
        return True

    def wired(self):
        for t in self.ts:
            if not t.wired():
                return False
        return True

    # test whether the values refer a whole tuple
    @staticmethod
    def _iswhole(ts, keys):

        for i in range(len(ts)):
            if ts[i].anon():
                #print('not an inst or defn')
                return False

        for i in range(len(ts)):
            # elements must be an tuple reference
            if not isinstance(ts[i].name, TupleRef):
                #print('not an tuple ref')
                return False

        for i in range(1,len(ts)):
            # elements must refer to the same tuple
            if ts[i].name.tuple is not ts[i-1].name.tuple:
                return False

        for i in range(len(ts)):
            # elements should be numbered consecutively
            if ts[i].name.index != list(keys)[i]:
                return False

        if len(ts) != len(ts[0].name.tuple):
            # elements should refer to a whole tuple
            return False

        return True

    def iswhole(self):
        return Tuple._iswhole(self.ts, self.keys())

    def trace(self, skip_self=True):
        ts = []
        for t in self.ts:
            result = t.trace(skip_self)
            if result is not None:
                ts.append(result)
            elif not skip_self and (t.is_output() or t.is_inout()):
                ts.append(t)
            else:
                return None

        if len(ts) == len(self) and Tuple._iswhole(ts, self.keys()):
            return ts[0].name.tuple

        return type(self).flip()(*ts)

    def value(self):
        ts = [t.value() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if len(ts) == len(self) and Tuple._iswhole(ts, self.keys()):
            return ts[0].name.tuple

        return type(self).flip()(*ts)

    def driving(self):
        return {k: t.driving() for k, t in self.items()}

    @classmethod
    def unflatten(cls, value):
        values = []
        offset = 0
        for field in cls.fields:
            size = field.flat_length()
            ts = value[offset:offset + size]
            values.append(field.unflatten(ts))
            offset += size
        return cls(*values)

    def flatten(self):
        return sum([t.flatten() for t in self.ts], [])

    def const(self):
        for t in self.ts:
            if not t.const():
                return False

        return True

    @classmethod
    def types(cls):
        return cls.fields

    def values(self):
        return self.ts

    def items(self):
        return zip(self.keys(), self.ts)

    def undriven(self):
        for elem in self:
            elem.undriven()

    def unused(self):
        for elem in self:
            elem.unused()

    @classmethod
    def is_mixed(cls):
        input, output, inout, mixed = False, False, False, False
        for field in cls.fields:
            input |= field.is_input()
            output |= field.is_output()
            inout |= field.is_inout()
            mixed |= field.is_mixed()
        return mixed or (input + output + inout) > 1

    def connection_iter(self, only_slice_bits=False):
        for elem in self:
            yield elem, elem.trace()

    def has_children(self):
        return True


def _add_properties(ns, fields):
    def _make_prop(field_type, idx):
        @TypedProperty(field_type)
        def prop(self):
            return self[idx]

        @prop.setter
        def prop(self, value):
            self[idx] = value

        return prop

    for idx, (field_name, field_type) in enumerate(fields.items()):
        if field_name == "N":
            # TODO: Make N unreserved
            raise ValueError("N is a reserved name in Product")
        assert field_name not in ns
        ns[field_name] = _make_prop(field_type, idx)


class AnonProductKind(AnonymousProductMeta, TupleKind, Kind):
    __hash__ = type.__hash__

    def _from_idx(cls, idx):
        mcs = type(cls)

        try:
            return mcs._class_cache[cls, idx]
        except KeyError:
            pass

        if cls.is_bound:
            raise TypeError('Type is already bound')

        undirected_idx = OrderedFrozenDict(
            [(k, v.qualify(Direction.Undirected)) for k, v in idx.items()]
        )
        if undirected_idx != idx:
            bases = [cls[undirected_idx]]
        else:
            bases = [cls]
        bases = tuple(bases)
        class_name = cls._name_from_idx(idx)

        ns = {'__module__' : cls.__module__}
        # add properties to namespace
        # build properties
        _add_properties(ns, idx)

        t = mcs(class_name, bases, ns, fields=tuple(idx.values()))
        t._field_table_ = idx
        if t._unbound_base_ is None:
            t._unbound_base_ = cls
        mcs._class_cache[cls, idx] = t
        t._cached_ = True
        return t

    def qualify(cls, direction):
        new_fields = OrderedDict()
        base = cls
        for k, v in cls.field_dict.items():
            new_fields[k] = v.qualify(direction)
        for k, v in cls.field_dict.items():
            if not issubclass(new_fields[k], v):
                base = cls.unbound_t
        if base.is_bound and all(v is base.field_dict[k]
                                 for k, v in new_fields.items()):
            return base

        if cls.unbound_t is AnonProduct:
            return cls.unbound_t._cache_handler(OrderedFrozenDict(new_fields))
        else:
            return cls.unbound_t._cache_handler(cls.is_cached, new_fields,
                                                cls.__name__, (base, ), {})

    def flip(cls):
        new_fields = OrderedDict()
        base = cls
        for k, v in cls.field_dict.items():
            new_fields[k] = v.flip()
        for k, v in cls.field_dict.items():
            if not issubclass(new_fields[k], v):
                base = cls.qualify(Direction.Undirected)
        if cls.unbound_t is AnonProduct:
            return cls.unbound_t._cache_handler(OrderedFrozenDict(new_fields))
        else:
            return cls.unbound_t._cache_handler(cls.is_cached, new_fields,
                                                cls.__name__, (base, ), {})

    def __eq__(cls, rhs):
        if not isinstance(rhs, AnonProductKind):
            return False

        if not cls.is_bound:
            return not rhs.is_bound

        return cls.field_dict == rhs.field_dict

    def is_wireable(cls, rhs):
        rhs = magma_type(rhs)
        if not isinstance(rhs, AnonProductKind) or len(cls.fields) != len(rhs.fields):
            return False
        for k, v in cls.field_dict.items():
            if not k in rhs.field_dict or not v.is_wireable(rhs.field_dict[k]):
                return False
        return True

    def is_bindable(cls, rhs):
        rhs = magma_type(rhs)
        if not isinstance(rhs, AnonProductKind) or len(cls.fields) != len(rhs.fields):
            return False
        for k, v in cls.field_dict.items():
            if not k in rhs.field_dict or not v.is_bindable(rhs.field_dict[k]):
                return False
        return True

    def __len__(cls):
        return len(cls.fields)

    def __str__(cls):
        if not cls.is_bound:
            return cls.__name__
        s = "Tuple("
        s += ",".join(f'{k}={str(v)}' for k, v in cls.field_dict.items())
        s += ")"
        return s


class AnonProduct(Tuple, metaclass=AnonProductKind):
    @classmethod
    def keys(cls):
        return cls.field_dict.keys()

    @classmethod
    def types(cls):
        return cls.fields

    def value(self):
        ts = [t.value() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if len(ts) == len(self) and Tuple._iswhole(ts, self.keys()):
            return ts[0].name.tuple

        return type(self).flip()(*ts)


class ProductKind(ProductMeta, AnonProductKind, Kind):
    __hash__ = type.__hash__

    def __new__(mcs, name, bases, namespace, cache=True, **kwargs):
        return super().__new__(mcs, name, bases, namespace, cache, **kwargs)

    def from_fields(cls, name, fields , cache=True):
        return super().from_fields(name, fields, cache)

    @classmethod
    def _from_fields(mcs, fields, name, bases, ns, **kwargs):

        cls = bases[0]

        # add properties to namespace
        # build properties
        _add_properties(ns, fields)

        # this is all really gross but I don't know how to do this cleanly
        # need to build t so I can call super() in new and init
        # need to exec to get proper signatures
        t = TupleKind.__new__(mcs, name, bases, ns, fields=tuple(fields.values()),
                              **kwargs)
        if t._unbound_base_ is None:
            t._unbound_base_ = cls

        # not strictly necessary could iterative over class dict finding
        # TypedProperty to reconstruct _field_table_ but that seems bad
        t._field_table_ = OrderedFrozenDict(fields)
        return t


class Product(AnonProduct, metaclass=ProductKind):
    pass



from .array import Array, ArrayType
from .conversions import array
from .circuit import AnonymousCircuit
from .wire import wire

__ALL__  = ['compose']
__ALL__ += ['curry', 'uncurry']
__ALL__ += ['flat', 'partition']
__ALL__ += ['row', 'col', 'map_']
__ALL__ += ['fork', 'join']
__ALL__ += ['fold', 'scan']
__ALL__ += ['braid']

# flatten list of lists
def flatten(l):
    return sum(l, [])

# return a list of all the arguments
# with a given name from a list of interfaces
def getarg(arg, interfaces):
    return [i.ports[arg] for i in interfaces]

#
def lscanarg(iarg, oarg, interfaces, noiarg=False, nooarg=False):
    iargs = getarg(iarg, interfaces)
    oargs = getarg(oarg, interfaces)
    n = len(interfaces)
    for i in range(n-1):
        wire(oargs[i], iargs[i+1])

    args = []
    if not noiarg:
        args += [iarg, iargs[0]]
    if not nooarg:
        args += [oarg, array(oargs)]
    return args

#
def rscanarg(iarg, oarg, interfaces, noiarg=False, nooarg=False):
    iargs = getarg(iarg, interfaces)
    oargs = getarg(oarg, interfaces)
    n = len(interfaces)
    for i in range(n-1):
        wire(oargs[i+1], iargs[i])

    args = []
    if not noiarg:
        args += [iarg, iargs[0]]
    if not nooarg:
        args += [oarg, array(oargs)]
    return args

#
def lfoldarg(iarg, oarg, interfaces, noiarg=False, nooarg=False):
    iargs = getarg(iarg, interfaces)
    oargs = getarg(oarg, interfaces)
    n = len(interfaces)
    for i in range(n-1):
        wire(oargs[i], iargs[i+1])
    args = []
    if not noiarg:
        args += [iarg, iargs[0]]
    if not nooarg:
        args += [oarg, oargs[n-1]]
    return args

def rfoldarg(iarg, oarg, interfaces, noiarg=False, nooarg=False):
    iargs = getarg(iarg, interfaces)
    oargs = getarg(oarg, interfaces)
    n = len(interfaces)
    for i in range(n-1):
        wire(oargs[i+1], iargs[i])
    args = []
    if not noiarg:
        args += [iarg, iargs[0]]
    if not nooarg:
        args += [oarg, oargs[n-1]]
    return args

# return [arg, args] from a list of interfaces
def forkarg(arg, interfaces):
    iargs = getarg(arg, interfaces)
    T = type(iargs[0])
    oarg = T.undirected_t() # create a single anonymous value
    for iarg in iargs:
        wire(oarg, iarg) # wire the anonymous value to all the forked args
    return [arg, oarg]

# return [arg, array] from a list of interfaces
def joinarg(arg, interfaces):
    args = getarg(arg, interfaces)
    #print('joinarg', args)
    return [arg, array(args)]


# all powerful braid functions
#
#  circuits : list of circuits which should all have the same signature
#
#  forkargs : list of argument names to fork
#   this argument is wired to all the circuits
#  joinargs : list of argument names to join
#   this argument is equal an array of all the arguments from circuits
#  foldargs : dict of argument namein:nameout, set namein[i+1] to namout[i]
#  rfoldargs : dict of argument namein:nameout, set namein[i-1] to namout[i]
#    namein[0] is retained in the result,
#    nameout[n-1] is retained in the result
#  scanargs : dict of argument namein:nameout, set namein[i+1] to namout[i]
#  rscanargs : dict of argument namein:nameout, set namein[i-1] to namout[i]
#    namein[0] is retained in the result,
#    the array nameout is retained in the result
#
# by default, clock arguments are forked,
#  unless they appear in another keyword
#
# by default, any arguments not appearing in a keyword are joined
#
def braid(circuits,
            forkargs=[],
            joinargs=[],
            foldargs={}, rfoldargs={},
            scanargs={}, rscanargs={}):

    forkargs = list(forkargs)

    interfaces = [circuit.interface for circuit in circuits]

    # by default, clkargs are added to forkargs,
    # unless they appear in another keyword
    for clkarg in interfaces[0].clockargnames():
        if clkarg in forkargs: continue
        if clkarg in joinargs: continue
        if clkarg in foldargs.keys(): continue
        if clkarg in rfoldargs.keys(): continue
        if clkarg in scanargs.keys(): continue
        if clkarg in rscanargs.keys(): continue
        forkargs.append(clkarg)

    # do NOT join arguments if they appear in another keyword
    nojoinargs = list(forkargs)
    nojoinargs += flatten( [[k, v] for k, v in foldargs.items()] )
    nojoinargs += flatten( [[k, v] for k, v in rfoldargs.items()] )
    nojoinargs += flatten( [[k, v] for k, v in scanargs.items()] )
    nojoinargs += flatten( [[k, v] for k, v in rscanargs.items()] )

    joinargs = [name for name in interfaces[0].ports.keys() \
                    if name not in nojoinargs]

    #print('fork', forkargs)
    #print('flat', flatargs)
    #print('join', joinargs)
    #print('fold', foldargs)
    #print('scan', scanargs)
    args = []
    for key in interfaces[0].ports.keys():
        if   key in foldargs:
            iarg = key
            oarg = foldargs[key]
            noiarg = iarg in forkargs or iarg in joinargs
            nooarg = oarg in joinargs
            #print('lfolding', iarg, key, noiarg, nooarg)
            args += lfoldarg(iarg, oarg, interfaces, noiarg, nooarg)
        elif key in rfoldargs:
            iarg = key
            oarg = rfoldargs[key]
            #print('rfolding', iarg, key)
            noiarg = iarg in forkargs or iarg in joinargs
            nooarg = oarg in joinargs
            args += rfoldarg(iarg, oarg, interfaces, noiarg, nooarg)

        elif key in scanargs:
            iarg = key
            oarg = scanargs[key]
            #print('scanning', iarg, key)
            noiarg = iarg in forkargs or iarg in joinargs
            nooarg = oarg in joinargs
            args += lscanarg(iarg, oarg, interfaces, noiarg, nooarg)
        elif key in rscanargs:
            iarg = key
            oarg = rscanargs[key]
            #print('scanning', iarg, key)
            noiarg = iarg in forkargs or iarg in joinargs
            nooarg = oarg in joinargs
            args += rscanarg(iarg, oarg, interfaces, noiarg, nooarg)

        elif key in forkargs:
            #print('forking', key)
            args += forkarg(key, interfaces)

        elif key in joinargs:
            #print('joining', key)
            args += joinarg(key, interfaces)

    #print(args)
    return AnonymousCircuit(args)

# fork all inputs
def fork(*circuits):
    """Wire input to all the inputs, concatenate output"""
    if len(circuits) == 1:
        circuits = circuits[0]
    return braid(circuits, forkargs=circuits[0].interface.inputs_by_name())

# join all inputs
def join(*circuits, joinargs=[]):
    """concatenate input and concatenate output"""
    if len(circuits) == 1:
        circuits = circuits[0]
    return braid(circuits)



def fold(*circuits, foldargs={"I":"O"}, **kwargs):
    """fold"""
    if len(circuits) == 1:
        circuits = circuits[0]
    return braid(circuits, foldargs=foldargs, **kwargs)

def scan(*circuits, scanargs={"I":"O"}, **kwargs):
    """scan"""
    if len(circuits) == 1:
        circuits = circuits[0]
    return braid(circuits, scanargs=scanargs, **kwargs)


def compose2(circuit1, circuit2):
    # collect input argument names from circuit1
    iargs = circuit1.interface.inputargs()
    iargs = [iargs[2*i] for i in range(len(iargs)//2)]

    # collect output argument names from circuit2
    oargs = circuit2.interface.outputargs()
    oargs = [oargs[2*i] for i in range(len(oargs)//2)]

    if len(iargs) != len(oargs):
        error("Number of inputs must equal the number of outputs")

    # wire outputs of circuit2 to the inputs of circuit1
    for I, O in zip(iargs, oargs):
        #print('wire({},{})'.format(O,I))
        wire( getattr(circuit2,O), getattr(circuit1,I) )

    # compute new arguments
    args = circuit2.interface.inputargs() + circuit1.interface.outputargs()

    # fork clock arguments
    clkargs1 = set(circuit1.interface.clockargnames())
    clkargs2 = set(circuit2.interface.clockargnames())
    if clkargs1 != clkargs2:
        error("Circuits have different clock arguments")
    interfaces = [circuit2.interface, circuit1.interface]
    for clkarg in clkargs1:
        args += forkarg(clkarg, interfaces)

    return AnonymousCircuit(args)

def compose(*circuits):
    if len(circuits) == 1:
        circuits = circuits[0]

    if len(circuits) == 1:
        return circuits[0]

    return compose2(circuits[0], compose(circuits[1:])) # right associative


#
# curry a circuit
#
#  the input argument named prefix, which must be an array,
#  is broken into separate input arguments
#
#  for example, if prefix='I',
#    then "I", array([i0, i1]) -> "I0", i0, "I1", i1
#
# all other inputs remain unchanged
#
def curry(circuit, prefix='I'):
    args = []
    for name, port in circuit.interface.ports.items():
        if not port.wired() and name == prefix and port.is_input():
           for i in range(len(port)):
               args.append('{}{}'.format(name, i))
               args.append(port[i])
        else:
           args.append(name)
           args.append(port)

    #print(args)
    return AnonymousCircuit(args)


def uncurry(circuit, prefix='I'):
    """

    uncurry a circuit

    all input arguments whose names begin with prefix
    are collected into a single input argument named prefix,
    which is an array constructed from of the input arguments

    for example, if prefix='I',
        then "I0", i0, "I1", i1 -> "I", array([i0, i1])

    the uncurry argument is the first argument in the result

    all other inputs remain unchanged.


    """
    otherargs = []
    uncurryargs = []
    for name, port in circuit.interface.ports.items():
        # should we insert the argument in the position of the first match?
        if not port.driven() and name.startswith(prefix) and \
                circuit.interface.is_input(port):
            uncurryargs.append(port)
        else:
            otherargs += [name, port]

    return AnonymousCircuit([prefix, array(uncurryargs)] + otherargs)

# flatten all the args matching names in flatargs
#  each input must be an array
def flat(circuit, flatargs=['I', 'O']):
    args = []
    for name, port in circuit.interface.ports.items():
        if not port.wired() and name in flatargs \
               and isinstance(port, ArrayType) \
                   and isinstance(port[0], ArrayType):
           #print('flat',name)
           ts = sum([p.as_list() for p in port], [])
           args += [name, array(ts)]
        else:
           args += [name, port]
    return AnonymousCircuit( args )

# concat all the inputs whose name starts with prefix
#  each input must be an array
def partition(circuit, n, prefix='I'):
    args = []
    for name, port in circuit.interface.ports.items():
        # should we insert the argument in the position of the first match?
        if not port.wired() and name == prefix and isinstance(port, ArrayType):
           l = port.as_list()
           l = [array(l[i:i + n]) for i in range(0, len(l), n)]
           for i in range(len(l)):
               args += ['{}{}'.format(prefix,i), l[i]]
        else:
           args += [name, port]
    return AnonymousCircuit( args )


def row(f, n):
    return [f(x) for x in range(n)]

def col(f, n):
    return [f(y) for y in range(n)]

def map_(f, n):
    return [f() for _ in range(n)]



from magma.circuit import peek_definition_context_stack
from magma.t import Type


class _Time:
    pass


def time():
    return _Time()


class _Event:
    def __init__(self, value):
        if not isinstance(value, Type):
            raise TypeError("Expected magma value for event")
        self.value = value


class _Posedge(_Event):
    verilog_str = "posedge"


class _Negedge(_Event):
    verilog_str = "negedge"


def posedge(value):
    return _Posedge(value)


def negedge(value):
    return _Negedge(value)


def _make_display_format_arg(value, format_args):
    # Unique name based on object id
    var = f"_display_var_{id(value)}"
    if isinstance(value, _Time):
        # Insert $time string
        value = "$time"
    # Insert into a format kwargs map
    format_args[var] = value
    # Wrap in format braces so it is replaced later on
    var = f"{{{var}}}"
    return var


class Display:
    def __init__(self, display_str, args, file):
        # Encode to handle newlines, etc... properly
        self.display_str = display_str.encode('unicode_escape').decode()
        self.args = args
        self.events = []
        self.cond = None
        self.file = file

    def if_(self, cond):
        """
        Method to set condition for display

            # Display if CE (enable) is high
            m.display("x=%d", x).when(m.posedge(io.CLK))\
                                .if_(io.CE)

        """
        if self.cond is not None:
            raise Exception("Can only invoke if_ once on display")
        self.cond = cond
        return self

    def when(self, event):
        """
        Allows chaining to set event for display, e.g.

            m.display("x=%d", x).when(m.posedge(io.CLK))\
                                .when(m.negedge(io.ASYNCRESET))

        """
        if not isinstance(event, (Type, _Event)):
            raise TypeError("Expected magma value or event for when argument")
        self.events.append(event)
        return self

    def _make_cond_str(self, format_args):
        if self.cond is not None:
            return _make_display_format_arg(self.cond, format_args)
        return ""

    def get_inline_verilog(self):
        format_args = {}
        display_args = []
        # arguments to the dipslay function are unique names that are
        # interpolated later by the inline_verilog syntax
        #     e.g. $display("...", {_display_var_0}, {_display_var_1});
        for arg in self.args:
            display_args.append(_make_display_format_arg(arg, format_args))

        display_args_str = ""
        if display_args:
            display_args_str = ", " + ", ".join(display_args)

        # Default all events
        event_str = "*"
        if self.events:
            event_strs = []
            for event in self.events:
                value = event
                # Could be sensitive to plain signal
                if isinstance(event, _Event):
                    value = value.value
                var = _make_display_format_arg(value, format_args)
                # prepend event if not just plain signal
                if isinstance(event, _Event):
                    var = f"{event.verilog_str} {var}"
                event_strs.append(var)
            event_str = ", ".join(event_strs)

        cond_str = self._make_cond_str(format_args)
        if cond_str:
            cond_str = f"if ({cond_str}) "

        display_str = f"$display("
        if self.file is not None:
            display_str = f"$fdisplay(\\_file_{self.file.filename} , "  # noqa
        format_str = f"""\
always @({event_str}) begin
    {cond_str}{display_str}\"{self.display_str}\"{display_args_str});
end
"""
        return format_str, format_args, {}


def display(display_str, *args, file=None):
    context = peek_definition_context_stack()
    disp = Display(display_str, args, file)
    context.get_child("display").add_display(disp)
    return disp


class File:
    def __init__(self, filename, mode):
        self.filename = filename
        self.mode = mode

        context = peek_definition_context_stack()
        context.get_child("display").add_file(self)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass



import inspect
import textwrap
import ast
import os
import types
import typing
import astor
import traceback
import functools

from ast_tools.stack import inspect_enclosing_env

def get_ast(obj):
    indented_program_txt = inspect.getsource(obj)
    program_txt = textwrap.dedent(indented_program_txt)
    return ast.parse(program_txt)

def get_func_ast(obj : types.FunctionType):
    """ Implicitly strip ast.Module() surrounding the function """
    return get_ast(obj).body[0]


def compile_function_to_file(tree : typing.Union[ast.Module, ast.FunctionDef],
                             func_name : str = None, defn_env : dict = None):
    if defn_env is None:
        defn_env = {}

    if isinstance(tree, ast.FunctionDef):
        if func_name is None:
            func_name = tree.name
        else:
            if func_name != tree.name:
                raise Exception("Passed in func_name that does not match the "
                                "function being compiled. Got"
                                f" func_name={func_name} expected"
                                f" tree.name={tree.name}")
    elif isinstance(tree, ast.Module):
        if func_name is None:
            raise Exception("func_name required when passing in an ast.Module")
    os.makedirs(".magma", exist_ok=True)
    file_name = os.path.join(".magma", func_name + ".py")
    with open(file_name, "w") as fp:
        fp.write(astor.to_source(tree))

    try:
        code = compile(astor.to_source(tree), filename=file_name, mode="exec")
    except:
        import sys
        tb = traceback.format_exc()
        print(tb)
        raise Exception(f"Error occured when compiling m.circuit.combinational function {func_name}, see above")

    try:
        exec(code, defn_env)
    except:
        import sys
        tb = traceback.format_exc()
        print(tb)
        raise Exception(f"Error occured when executing m.circuit.combinational function {func_name}, see above")

    return defn_env[func_name]

class NameCollector(ast.NodeVisitor):
    def __init__(self):
        self.names = set()

    def visit_Name(self, node: ast.Name):
        self.names.add(node.id)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.names.add(node.name)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        self.names.add(node.name)

    def visit_ClassDef(self, node: ast.ClassDef):
        self.names.add(node.name)

def gen_free_name(tree: ast.AST, defn_env: dict, prefix: str = '__auto_name_'):
    visitor = NameCollector()
    visitor.visit(tree)
    used_names = visitor.names | defn_env.keys()
    f_str = prefix+'{}'
    c = 0
    name = f_str.format(c)
    while name in used_names:
        c += 1
        name = f_str.format(c)

    return name

def filter_decorator(decorator : typing.Callable, decorator_list : typing.List[ast.AST], env : dict):
    def _filter(node):
        if isinstance(node, ast.Call):
            expr = ast.Expression(node.func)
        elif isinstance(node, ast.Name):
            expr = ast.Expression(node)
        else:
            return True
        code = compile(expr, filename="<string>", mode="eval")
        e = eval(code, env)
        return e != decorator
    return list(filter(_filter, decorator_list))



from magma import *
import coreir

context = coreir.Context()

def get_lib(lib):
    if lib in {"coreir", "mantle", "corebit", "memory"}:
        return context.get_namespace(lib)
    elif lib == "global":
        return context.global_namespace
    else:
        return context.load_library(lib)

def import_generator(lib, name):
    generator = get_lib(lib).generators[name]
    def _generator(**kwargs):
        return import_circuit(generator(**kwargs))
    return _generator

def to_magma_type(typ):
    if typ.kind == "Array":
        return Array(len(typ), to_magma_type(typ.element_type))
    elif typ.kind == "BitIn":
        return In(Bit)
    elif typ.kind == "Bit":
        return Out(Bit)
    raise ValueError(f"Cannot convert type: {typ.kind}")

def import_circuit(module):
    if not isinstance(module, coreir.Module):
        raise ValueError(f"Excepted module, not {module}")
    io = []
    for name, typ in module.type.items():
        io += [name, to_magma_type(typ)]
    if module.definition is None:
        circ = DeclareCircuit(module.name, *io)
    else:
        circ = DefineCircuit(module.name, *io)
        raise NotImplementedError()
    return circ



import dataclasses
from typing import List, Optional, Sequence, Tuple

from magma.symbol_table import (SymbolTableInterface, SymbolTable,
                                DelegatorSymbolTable, ImmutableSymbolTable,
                                SYMBOL_TABLE_INLINED_INSTANCE,
                                SYMBOL_TABLE_EMPTY)


@dataclasses.dataclass
class _Object:
    name: str
    scope: int
    root: Optional['_Object'] = None

    def __post_init__(self):
        self._renames: List[_Rename] = []

    @property
    def renames(self):
        return self._renames.copy()

    def root_or_self(self):
        if self.root is not None:
            return self.root
        return self

    def tail(self):
        if not self._renames:
            return self
        if len(self._renames) == 1:
            return self._renames[0].obj.tail()
        raise ValueError(self)

    def key(self):
        return (self.scope, self.name)

    def add_rename(self, rename: '_Rename'):
        self._renames.append(rename)


@dataclasses.dataclass
class _Rename:
    obj: _Object
    modifiers: Tuple[str] = ()


@dataclasses.dataclass
class _Module(_Object):
    def __post_init__(self):
        super().__post_init__()
        self._instances: List['_Instance'] = {}
        self._ports: List['_Port'] = {}

    @property
    def instances(self):
        return self._instances.values()

    @property
    def ports(self):
        return self._ports.values()

    def get_rename(self):
        if len(self._renames) != 1:
            raise ValueError(self)
        return self._renames[0]

    def add_instance(self, instance: '_Instance'):
        assert instance.name not in self._instances
        self._instances[instance.name] = instance

    def has_instance(self, name: str):
        return name in self._instances

    def get_instance(self, name: str):
        return self._instances[name]

    def add_port(self, port: '_Port'):
        assert port.name not in self._ports
        self._ports[port.name] = port

    def get_port(self, name: str):
        return self._ports[name]

    def remove_port(self, name: str):
        del self._ports[name]


@dataclasses.dataclass
class _Instance(_Object):
    type: Optional[_Module] = None

    def __post_init__(self):
        super().__post_init__()
        self._inlines = None

    @property
    def inlined(self):
        return self._inlines is not None

    @property
    def inlines(self):
        if not self.inlined:
            return None
        return self._inlines.copy()

    def inline(self):
        if self.inlined:
            raise ValueError(self)
        self._inlines = []

    def add_inline(self, original, new):
        self._inlines.append((original, new))


@dataclasses.dataclass(eq=True, frozen=True)
class _PortWrapper:
    base: str
    fields: Tuple[str]

    def longname(self):
        if not self.fields:
            return self.base
        return self.base + "." + ".".join(self.fields)

    @staticmethod
    def make(longname: str):
        parts = longname.split(".")
        if not len(parts) > 0:
            raise ValueError(longname)
        if len(parts) == 1:
            return _PortWrapper(parts[0], ())
        return _PortWrapper(parts[0], tuple(parts[1:]))


@dataclasses.dataclass
class _Port(_Object):
    def __post_init__(self):
        super().__post_init__()
        self._wrapper = _PortWrapper.make(self.name)

    @property
    def wrapper(self):
        return self._wrapper

    def tails(self):
        if not self._renames:
            return [(self, ())]
        tails = []
        for rename in self._renames:
            sub_tails = rename.obj.tails()
            for obj, modifiers in sub_tails:
                tails.append((obj, rename.modifiers + modifiers))
        return tails


def _finalize_instance_names(module, instance, table, top=True):
    tail = instance.tail()
    if top:
        if not tail.inlined:
            table.set_instance_name(
                module.name, instance.name, (SYMBOL_TABLE_EMPTY, tail.name))
            return
        table.set_instance_name(
            module.name, instance.name, (SYMBOL_TABLE_INLINED_INSTANCE, ""))
    for original, new in tail.inlines:
        if not new.inlined:
            table.set_inlined_instance_name(
                module.name,
                instance.name,
                original.root.name,
                (SYMBOL_TABLE_EMPTY, new.tail().name))
        else:
            table.set_inlined_instance_name(
                module.name,
                instance.name,
                original.root.name,
                (SYMBOL_TABLE_INLINED_INSTANCE, new.tail().name))
            _finalize_instance_names(module, new, table, top=False)


class _TableProcessor:
    def __init__(self):
        self._modules = {}
        self._scope = 0
        self._finalized = False
        self._uniq_key_map = {}
        self._uniq_key_counter = 0

    def process_table(self, table: SymbolTableInterface):
        self._process_module_names(table.module_names())
        self._process_instance_names(table.instance_names())
        self._process_port_names(table.port_names())
        self._process_instance_types(table.instance_types())
        self._process_inlined_instance_names(table.inlined_instance_names())
        self._scope += 1

    def finalize(self, table: SymbolTableInterface):
        if self._finalized:
            raise Exception("Can not call finalize() multiple times")
        root_modules = list(
            filter(lambda m: m.scope == 0, self._modules.values()))
        for module in root_modules:
            table.set_module_name(module.name, module.tail().name)
            for instance in module.instances:
                _finalize_instance_names(module, instance, table)
                table.set_instance_type(
                    module.name, instance.name, instance.type.name)
            for port in module.ports:
                for tail, modifiers in port.tails():
                    src_port = _PortWrapper(port.name, modifiers).longname()
                    table.set_port_name(module.name, src_port, tail.name)
        self._finalized = True

    def _new_unique_key(self):
        key = f"UNIQ_KEY_{self._uniq_key_counter}"
        self._uniq_key_counter += 1
        return key

    def _get_or_set_uniq_key(self, scope, uniq_key):
        dict_key = (scope, uniq_key)
        try:
            mapped = self._uniq_key_map[dict_key]
        except KeyError:
            mapped = self._new_unique_key()
            self._uniq_key_map[dict_key] = mapped
        return mapped

    def _process_module_names(self, module_names):
        if self._scope != 0:
            unmapped_srcs = {m.name: m
                             for m in self._modules.values()
                             if m.scope == self._scope}
        for in_module_name, out_module_name in module_names.items():
            if self._scope == 0:
                src = _Module(in_module_name, self._scope)
                root = src
                self._modules[src.key()] = src
            else:
                # If this is a new module in this scope, then skip it.
                try:
                    src = self._modules[(self._scope, in_module_name)]
                except KeyError:
                    continue
                del unmapped_srcs[src.name]
                root = src.root
            dst = _Module(out_module_name, self._scope + 1, root=root)
            src.add_rename(_Rename(dst))
            self._modules[dst.key()] = dst
        if self._scope != 0:
            for src in unmapped_srcs.values():
                dst = _Module(src.name, self._scope + 1, root=src.root)
                src.add_rename(_Rename(dst))
                self._modules[dst.key()] = dst

    def _process_instance_names(self, instance_names):
        for key, value in instance_names.items():
            sentinel, out_instance_name = value
            if not (sentinel is SYMBOL_TABLE_EMPTY or
                    sentinel is SYMBOL_TABLE_INLINED_INSTANCE):
                raise ValueError((key, value))
            in_module_name, in_instance_name = key
            # If this is a new module in this scope, then skip it, unless this
            # is the base scope.
            try:
                src_module = self._modules[(self._scope, in_module_name)]
            except KeyError:
                if self._scope == 0:
                    raise
                continue
            dst_module = src_module.get_rename().obj
            if self._scope == 0:
                src_instance = _Instance(in_instance_name, self._scope)
                root_instance = src_instance
                src_module.add_instance(src_instance)
            else:
                if not src_module.has_instance(in_instance_name):
                    continue
                src_instance = src_module.get_instance(in_instance_name)
                root_instance = src_instance.root
            if sentinel is SYMBOL_TABLE_INLINED_INSTANCE:
                src_instance.inline()
            else:
                dst_instance = _Instance(
                    out_instance_name, self._scope + 1, root=root_instance)
                dst_module.add_instance(dst_instance)
                src_instance.add_rename(_Rename(dst_instance))

    def _process_port_names(self, port_names):
        for key, out_port_name in port_names.items():
            in_module_name, in_port_name = key
            # If this is a new module in this scope, then skip it.
            try:
                src_module = self._modules[(self._scope, in_module_name)]
            except KeyError:
                if self._scope == 0:
                    raise
                continue
            dst_module = src_module.get_rename().obj
            modifiers = ()
            if self._scope == 0:
                src_port = _Port(in_port_name, self._scope)
                root_port = src_port
                src_module.add_port(src_port)
            else:
                try:
                    src_port = src_module.get_port(in_port_name)
                except KeyError:
                    dummy_port = _Port(in_port_name, self._scope)
                    src_port = src_module.get_port(dummy_port.wrapper.base)
                    modifiers = dummy_port.wrapper.fields
                root_port = src_port.root
            dst_port = _Port(out_port_name, self._scope + 1, root=root_port)
            dst_module.add_port(dst_port)
            src_port.add_rename(_Rename(dst_port, modifiers))

    def _process_instance_types(self, instance_types):
        for key, out_type in instance_types.items():
            module_name, instance_name = key
            # If this is a new module in this scope, then skip it.
            try:
                module = self._modules[(self._scope, module_name)]
            except KeyError:
                continue
            # If this is a new instance in this module (decl -> defn) then skip
            # it.
            if not module.has_instance(instance_name):
                continue
            instance = module.get_instance(instance_name)
            module_type = self._modules[(self._scope, out_type)]
            instance.type = module_type

    def _process_inlined_instance_names(self, inlined_instance_names):
        items = list(inlined_instance_names.items())
        while items:
            key, value = items.pop()
            module_name, parent_instance_name, child_instance_name = key
            sentinel, new_instance_name_or_key = value
            if not (sentinel is SYMBOL_TABLE_EMPTY or
                    sentinel is SYMBOL_TABLE_INLINED_INSTANCE):
                raise ValueError((key, value))
            src_module = self._modules[(self._scope, module_name)]
            dst_module = src_module.get_rename().obj
            parent_instance_name = self._uniq_key_map.get(
                (self._scope, parent_instance_name), parent_instance_name)
            try:
                parent_instance = src_module.get_instance(parent_instance_name)
            except KeyError:
                items.insert(0, (key, value))
                continue
            child_instance = parent_instance.type.get_instance(
                child_instance_name)
            if not parent_instance.inlined:
                raise ValueError((key, value))
            if sentinel is SYMBOL_TABLE_EMPTY:
                new_instance = _Instance(
                    new_instance_name_or_key, self._scope + 1)
                dst_module.add_instance(new_instance)
            else:
                new_instance_name_or_key = self._get_or_set_uniq_key(
                    self._scope, new_instance_name_or_key)
                new_instance = _Instance(new_instance_name_or_key, self._scope)
                new_instance.inline()
                src_module.add_instance(new_instance)
            new_instance.type = child_instance.type
            parent_instance.add_inline(child_instance, new_instance)


class MasterSymbolTable(ImmutableSymbolTable, DelegatorSymbolTable):
    def __init__(self, tables: Sequence[SymbolTableInterface]):
        super().__init__(SymbolTable())
        processor = _TableProcessor()
        for table in tables:
            processor.process_table(table)
        processor.finalize(self._underlying)


class SymbolQueryInterface:

    class InlinedLeafInstanceError(Exception):
        pass

    @dataclasses.dataclass
    class _Inst:
        name: str
        type: str

    def __init__(self, table: SymbolTableInterface):
        self._table = table

    def get_module_name(self, module_name):
        return self._table.get_module_name(module_name)

    def _get_inlined_instance_name_impl(
            self, module: str, parent_inst: _Inst, insts: List[_Inst]):
        if not insts:
            return []
        child_inst = insts[0]
        sentinel, new_name_or_key = self._table.get_inlined_instance_name(
            module, parent_inst.name, child_inst.name)
        if sentinel is SYMBOL_TABLE_EMPTY:
            name = new_name_or_key
            insts = insts[1:]
            return [name] + self._get_instance_name_impl(child_inst.type, insts)
        if sentinel is SYMBOL_TABLE_INLINED_INSTANCE:
            insts = insts[1:]
            if not insts:
                raise SymbolQueryInterface.InlinedLeafInstanceError()
            key = SymbolQueryInterface._Inst(new_name_or_key, None)
            return self._get_inlined_instance_name_impl(module, key, insts)
        raise ValueError(sentinel)

    def _get_instance_name_impl(self, module: str, insts: List[_Inst]):
        if not insts:
            return []
        inst = insts[0]
        sentinel, name = self._table.get_instance_name(module, inst.name)
        if sentinel is SYMBOL_TABLE_EMPTY:
            return [name] + self._get_instance_name_impl(inst.type, insts[1:])
        if sentinel is SYMBOL_TABLE_INLINED_INSTANCE:
            insts = insts[1:]
            if not insts:
                raise SymbolQueryInterface.InlinedLeafInstanceError()
            return self._get_inlined_instance_name_impl(module, inst, insts)
        raise ValueError(sentinel)

    def _get_types(self, module: str, insts: List[str]):
        curr_module = module
        out = []
        for inst in insts:
            curr_module = self._table.get_instance_type(curr_module, inst)
            out.append(curr_module)
        return out

    def get_instance_name(self, path: str) -> Optional[str]:
        module, *insts = path.split(".")
        if not insts:
            raise ValueError(path)
        insts = [SymbolQueryInterface._Inst(inst, t)
                 for inst, t in zip(insts, self._get_types(module, insts))]
        new_path = [self._table.get_module_name(module)]
        inst_path = self._get_instance_name_impl(module, insts)
        new_path += inst_path
        return ".".join(new_path)


def make_master_symbol_table(json_filenames: Sequence[str]):
    # TODO(rsetaluri): Make this lazy.
    symbol_tables = []
    for filename in json_filenames:
        with open(filename, "r") as f:
            table = SymbolTable.from_json(f.read())
            symbol_tables.append(table)
    return MasterSymbolTable(symbol_tables)



import functools
import libcst as cst
import libcst.matchers as match
from typing import Optional, MutableMapping

from ast_tools.common import gen_free_name
from ast_tools.passes import (apply_passes, ssa, if_to_phi, bool_to_bit, Pass,
                              PASS_ARGS_T)
from ast_tools.stack import SymbolTable

from ..circuit import Circuit, IO
from ..protocol_type import MagmaProtocol
from ..t import Type

from .util import build_io_args, build_call_args, wire_call_result
from magma.set_name import set_name


class NameSetter(cst.CSTTransformer):
    def __init__(self, set_name_var):
        self.set_name_var = set_name_var

    def leave_Assign(self, original_node, updated_node):
        if match.matches(updated_node,
                         match.Assign([match.AssignTarget(match.Name())])):
            return updated_node.with_changes(
                value=cst.Call(
                    cst.Name(self.set_name_var),
                    [cst.Arg(cst.SimpleString(
                        "\"" + updated_node.targets[0].target.value + "\"")
                    ), cst.Arg(updated_node.value)])
            )
        return updated_node


class insert_set_name(Pass):
    def rewrite(
        self, tree: cst.CSTNode, env: SymbolTable, metadata: MutableMapping
    ) -> PASS_ARGS_T:
        set_name_var = gen_free_name(tree, env)
        env.locals[set_name_var] = set_name
        return tree.visit(NameSetter(set_name_var)), env, metadata


class combinational2(apply_passes):
    def __init__(self, pre_passes=[], post_passes=[],
                 debug: bool = False,
                 env: Optional[SymbolTable] = None,
                 path: Optional[str] = None,
                 file_name: Optional[str] = None
                 ):
        passes = (pre_passes +
                  [ssa(strict=False), bool_to_bit(),
                   if_to_phi(lambda s, t, f: s.ite(t, f)),
                   insert_set_name()] +
                  post_passes)
        super().__init__(passes=passes, env=env, debug=debug, path=path,
                         file_name=file_name)

    def exec(self, *args, **kwargs):
        fn = super().exec(*args, **kwargs)

        io_args = build_io_args(fn.__annotations__)

        class CombinationalCircuit(Circuit):
            name = fn.__name__
            io = IO(**io_args)

            call_args = build_call_args(io, fn.__annotations__)

            call_result = fn(*call_args)

            wire_call_result(io, call_result, fn.__annotations__)

        @functools.wraps(fn)
        def func(*args, **kwargs):
            if (len(args) + len(kwargs) and
                not any(isinstance(x, (Type, MagmaProtocol)) for x in args +
                        tuple(kwargs.values()))):
                # If not called with at least one magma value, use the Python
                # implementation
                return fn(*args, **kwargs)
            return CombinationalCircuit()(*args, **kwargs)
        func.__name__ = fn.__name__
        func.__qualname__ = fn.__name__
        # Provide a mechanism for accessing the underlying circuit definition
        setattr(func, "circuit_definition", CombinationalCircuit)
        return func
